
import uuid
from collections import defaultdict
from textwrap import dedent

from django.db import transaction
from django.db.models import Count, Q
from drf_spectacular.utils import extend_schema_field
from rest_framework import permissions, serializers
from rest_framework.exceptions import PermissionDenied, ValidationError
from rest_framework.validators import UniqueTogetherValidator

from arkindex.documents.models import Element
from arkindex.documents.serializers.elements import ElementListSerializer
from arkindex.ponos.models import Task
from arkindex.process.models import Worker
from arkindex.project.serializer_fields import ArchivedField, DatasetSetsCountField, EnumField
from arkindex.training.models import (
    Dataset,
    DatasetElement,
    DatasetSet,
    DatasetState,
    MetricKey,
    MetricMode,
    MetricValue,
    Model,
    ModelVersion,
    ModelVersionState,
)
from arkindex.users.models import Role
from arkindex.users.utils import get_max_level


def _model_from_context(serializer_field):
    if isinstance(serializer_field.parent.instance, ModelVersion):
        return serializer_field.parent.instance.model
    return serializer_field.context.get("model")


def _corpus_from_context(serializer_field):
    if isinstance(serializer_field.parent.instance, Dataset):
        return serializer_field.parent.instance.corpus
    return serializer_field.context.get("corpus")


def _dataset_from_context(serializer_field):
    return serializer_field.context.get("dataset")


_model_from_context.requires_context = True
_corpus_from_context.requires_context = True
_dataset_from_context.requires_context = True


class ModelLightSerializer(serializers.ModelSerializer):
    name = serializers.CharField(max_length=100)

    class Meta:
        model = Model
        fields = ("id", "name")

    def validate_name(self, name):
        """Ensure model's unique name is available"""
        # Name unicity check is performed at the view level for creation
        if not self.instance:
            return name
        if Model.objects.using("default").filter(name=name).exclude(id=self.instance.id).exists():
            raise ValidationError("A model with this name already exists")
        return name


class ModelSerializer(ModelLightSerializer):
    # Actually define the field to avoid the field-level automatically generated UniqueValidator
    rights = serializers.SerializerMethodField(read_only=True)

    archived = ArchivedField(
        required=False,
        help_text=dedent("""
            Whether this model is archived.
            Model versions cannot be created on archived models, and they cannot be used in processes.

            Archiving or unarchiving an existing model requires admin access to the model.
        """),
    )

    class Meta(ModelLightSerializer.Meta):
        fields = ModelLightSerializer.Meta.fields + ("created", "updated", "archived", "description", "rights")

    def create(self, validated_data):
        instance = super().create(validated_data)

        # Create admin membership for creator user
        instance.memberships.create(
            user=self.context["request"].user,
            level=Role.Admin.value
        )
        return instance

    @extend_schema_field(serializers.ListField(child=serializers.ChoiceField(["read", "write", "admin"])))
    def get_rights(self, model):
        # to avoid a stale read on get_max_level when creating a model (because the corresponding membership is
        # created at the same time as the model) if a model is being created we return these rights without
        # actually checking them.
        if self.context["request"].method == "POST":
            return ["read", "write", "admin"]

        level = get_max_level(self.context["request"].user, model)
        rights = ["read"]
        if level >= Role.Contributor.value:
            rights.append("write")
        if level >= Role.Admin.value:
            rights.append("admin")
        return rights


class ModelCreateSerializer(ModelSerializer):
    # Exclude archived from model creation, since creating an archived model makes no sense
    archived = ArchivedField(read_only=True)


class CreateModelErrorResponseSerializer(serializers.Serializer):
    id = serializers.UUIDField(required=False, help_text="UUID of an existing model, if the error comes from a duplicate name.")
    name = serializers.CharField(required=False, help_text="Name of an existing model, if the error comes from a duplicate name.")
    detail = serializers.CharField(required=False, help_text="A generic error message when an error occurs outside of a specific field.")


class ModelCompatibleWorkerSerializer(serializers.ModelSerializer):
    model = serializers.PrimaryKeyRelatedField(queryset=Model.objects.none())
    worker = serializers.PrimaryKeyRelatedField(queryset=Worker.objects.none())

    class Meta:
        model = Model.compatible_workers.through
        fields = ("model", "worker")
        # Disable the implicit UniqueTogetherValidator that DRF adds,
        # as it will also run for deletions and always fail
        validators = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.context.get("request"):
            user = self.context["request"].user
            # Editing the compatible workers is editing an attribute of a model
            # Avoid a stale read if we just created a model and and want to add compatible workers to it
            self.fields["model"].queryset = Model.objects.editable(user).using("default")
            # Any worker that the user can browse can be set
            self.fields["worker"].queryset = Worker.objects.executable(user)

    def validate_model(self, model):
        if model.archived:
            raise ValidationError("This model is archived.")
        return model

    def validate(self, data):
        data = super().validate(data)
        # Ignore the uniqueness check for deletions
        if self.context["request"].method == "DELETE":
            return data

        # Avoid a stale read if creation is called multiple times in a row
        if data["model"].compatible_workers.filter(id=data["worker"].id).using("default").exists():
            raise ValidationError({"worker": ["This worker is already marked as compatible with this model."]})

        return data


class ModelVersionLightSerializer(serializers.ModelSerializer):
    """
    Serialize a simple model version
    """

    model = ModelLightSerializer(default=_model_from_context)
    state = EnumField(ModelVersionState)
    configuration = serializers.JSONField(style={"base_template": "textarea.html"})

    class Meta:
        model = ModelVersion
        fields = ("id", "model", "tag", "state", "size", "configuration")
        read_only_fields = ("id", "model", "state", "size", "hash", "archive_hash")


class ModelVersionSerializer(serializers.ModelSerializer):
    """
    Serialize a model version with fields that can be updated regardless of its state.
    """
    model = serializers.HiddenField(default=_model_from_context)
    parent = serializers.PrimaryKeyRelatedField(
        queryset=ModelVersion.objects.none(),
        default=None,
        allow_null=True,
    )
    description = serializers.CharField(required=False, style={"base_template": "textarea.html"})
    configuration = serializers.JSONField(required=False, decoder=None, encoder=None, style={"base_template": "textarea.html"})
    tag = serializers.CharField(allow_null=True, max_length=50, required=False, default=None)
    state = EnumField(ModelVersionState, read_only=True)
    s3_url = serializers.SerializerMethodField(read_only=True, help_text="Only returned if the user has **contributor** access to the model, and the model is available.")
    s3_put_url = serializers.SerializerMethodField(read_only=True, help_text="Only returned if the user has **contributor** access to the model, and the model is **not** available.")

    class Meta:
        model = ModelVersion
        fields = ("id", "model", "model_id", "parent", "description", "tag", "state", "size", "hash", "configuration", "s3_url", "s3_put_url", "created")
        read_only_fields = ("id", "model", "state", "size", "hash", "archive_hash")
        validators = [
            UniqueTogetherValidator(
                queryset=ModelVersion.objects.filter(tag__isnull=False),
                fields=["model", "tag"],
                message="A version for this model with this tag already exists."
            )
        ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if isinstance(self.instance, ModelVersion):
            model = self.instance.model
        else:
            model = self.context.get("model")
        if model:
            qs = ModelVersion.objects.filter(model__in=Model.objects.readable(self.context["request"].user))
            if getattr(self.instance, "id", None):
                qs = qs.exclude(id=self.instance.id)
            self.fields["parent"].queryset = qs

    def validate(self, data):
        model = data.get("model")
        if not model:
            model = self.instance.model

        if model.archived:
            raise ValidationError({"model": ["This model is archived."]})

        return data

    @extend_schema_field(serializers.CharField(allow_null=True))
    def get_s3_put_url(self, obj):
        if not self.context.get("is_contributor") or obj.state == ModelVersionState.Available:
            return None
        return obj.s3_put_url

    @extend_schema_field(serializers.CharField(allow_null=True))
    def get_s3_url(self, obj):
        if not self.context.get("is_contributor") or obj.state != ModelVersionState.Available:
            return None
        return obj.s3_url


class ModelVersionValidateSerializer(serializers.ModelSerializer):
    """
    A serializer used to update a model version to `available` state.
    """
    class Meta:
        model = ModelVersion
        fields = ("size", "hash", "archive_hash")
        extra_kwargs = {
            "size": {"required": True},
            "hash": {"required": True},
            "archive_hash": {"required": True},
        }


class MetricValueSerializer(serializers.ModelSerializer):
    class Meta:
        model = MetricValue
        fields = ("value", "created", "step")


class MetricValueCreateSerializer(MetricValueSerializer):
    """
    Create a new metric value, and the associated metric key if it doesn't already exist
    """
    model_version_id = serializers.PrimaryKeyRelatedField(
        queryset=ModelVersion.objects.select_related("model"),
        source="model_version",
        write_only=True
    )
    name = serializers.CharField(write_only=True)
    mode = EnumField(MetricMode, allow_null=True, required=False, write_only=True, default=MetricMode.Series)

    class Meta:
        model = MetricValue
        fields = MetricValueSerializer.Meta.fields + ("model_version_id", "name", "mode")

    def validate_model_version_id(self, model_version):
        # Assert user has a contributor access to the model
        access_level = get_max_level(self.context["request"].user, model_version.model)
        if not access_level or access_level < Role.Contributor.value:
            raise PermissionDenied(detail="You do not have contributor access to this model.")

        if model_version.model.archived:
            raise ValidationError("This ModelVersion is part of an archived model.")

        return model_version

    def validate(self, data):
        errors = defaultdict(list)
        data = super().validate(data)
        metric_value = {}
        name = data.get("name")
        # Because the mode defaults to MetricMode.Series if it is not set in the request, we need to retrieve what was
        # originally sent (or not) in the request to validate; otherwise if a request tries to create a metric value for
        # a metric key of type point, without setting mode, then data.get('mode') would be 'series' which would trigger
        # the metric_key.mode != mode validation error.
        mode = self.initial_data.get("mode")
        step = data.get("step")
        value = data.get("value")
        model_version = data.get("model_version")

        # Retrieve or create the metric key.
        # Annotating the Metric Key with the count of associated Metric Values (as metric_values_count) in order to check
        # for duplicate values for point mode metrics, and only check for duplicate values on series mode metrics for the
        # same step if there are existing values for this Metric Key.
        # Also annotating with the count of values with step set for the Metric Key, in order to check that a metric value
        # with step set isn't being sent for a Metric Key that already has step-less values, and vice-versa.
        queryset = MetricKey.objects \
            .annotate(metric_values_count=Count("values")) \
            .annotate(values_with_steps=Count("values", filter=Q(values__step__isnull=False)))
        # If step is being sent in the request, annotate with the count of values already set for that metric key for that
        # step, in order to check that no duplicate is being sent.
        if step is not None:
            queryset = queryset.annotate(current_step_values=Count("values", filter=Q(values__step=step)))
        metrickey, created = queryset \
            .get_or_create(name=name, model_version_id=model_version.id, defaults={"mode": data.get("mode")})

        # Check that the user isn't trying to set `step` on a point type metric key, which doesn't make sense as steps are
        # only relevant for series type metric values.
        if metrickey.mode == MetricMode.Point and step is not None:
            errors["step"].append("The step property cannot be set for Point type metrics.")

        # Check that the user isn't trying to change the metric key's mode, or adding a value to a point type metric value
        # if the metric key already exists.
        if not created:
            if mode and mode != metrickey.mode:
                errors["mode"].append(f"Metric Key {metrickey.name} already exists and is of mode {str(metrickey.mode)}.")
            if metrickey.mode == MetricMode.Point and metrickey.metric_values_count > 0:
                errors["__all__"].append(f"Metric Key {metrickey.name} is of mode Point and already has a value.")
            if metrickey.metric_values_count > 0 and metrickey.mode == MetricMode.Series and step is not None:
                if metrickey.current_step_values > 0:
                    errors["step"].append("This metric has already been set for this step.")
            # One metric key cannot have values both with and without step set
            if metrickey.metric_values_count > 0 and metrickey.mode == MetricMode.Series:
                if metrickey.values_with_steps > 0 and step is None:
                    errors["step"].append(f"Cannot add a metric value without a step for metric key {metrickey.name} "
                                          "because it already has metric values with steps set.")
                elif metrickey.values_with_steps == 0 and step is not None:
                    errors["step"].append(f"Cannot add a metric value with step set for metric key {metrickey.name} "
                                          "because it already has metric values with no step set.")

        metric_value["metric"] = metrickey
        metric_value["step"] = step
        metric_value["value"] = value

        if errors:
            raise ValidationError(errors)

        return metric_value


class MetricValueBulkItemSerializer(serializers.Serializer):
    name = serializers.CharField(write_only=True)
    mode = EnumField(MetricMode, allow_null=True, required=False, write_only=True, default=MetricMode.Series)
    value = serializers.FloatField()


class MetricValueBulkSerializer(serializers.Serializer):
    model_version_id = serializers.PrimaryKeyRelatedField(
        queryset=ModelVersion.objects.select_related("model").all(),
        source="model_version",
        write_only=True,
    )
    step = serializers.IntegerField(min_value=0, required=False)
    metrics = MetricValueBulkItemSerializer(
        many=True,
        allow_empty=False,
        help_text=dedent("""
        A list of metric values with their metric key name and mode.

        Only one value may be set at once for each metric key.
        """),
    )

    def validate_model_version_id(self, model_version):
        # Assert user has a contributor access to the model
        access_level = get_max_level(self.context["request"].user, model_version.model)
        if not access_level or access_level < Role.Contributor.value:
            raise PermissionDenied(detail="You do not have contributor access to this model.")

        if model_version.model.archived:
            raise ValidationError("This ModelVersion is part of an archived model.")

        return model_version

    def validate(self, data):
        errors = defaultdict(lambda: defaultdict(list))
        data = super().validate(data)
        model_version = data.get("model_version")
        step = data.get("step")

        # Retrieve the existing Metric Keys corresponding to the request metric names
        metric_names = set([item["name"] for item in data["metrics"]])
        queryset = MetricKey.objects \
            .annotate(metric_values_count=Count("values")) \
            .annotate(values_with_steps=Count("values", filter=Q(values__step__isnull=False))) \
            .filter(name__in=metric_names, model_version=model_version)
        if step is not None:
            queryset = queryset.annotate(current_step_values=Count("values", filter=Q(values__step=step)))
        metrickeys = {
            item.name: item
            for item in queryset
        }

        # Validate each Metric Value
        metric_names_list = []
        for i, item in enumerate(data["metrics"]):
            name = item["name"]
            if name not in metric_names_list:
                metric_names_list.append(name)
            else:
                errors[i]["__all__"].append(f"Cannot send multiple values at once for the same name ({name}).")
            # Because the mode defaults to MetricMode.Series if it is not set in the request, we need to retrieve what was
            # originally sent (or not) in the request to validate; otherwise if a request tries to create a metric value for
            # a metric key of type point, without setting mode, then data['metrics'][i].get('mode') would be 'series' which
            # would trigger the metric_key.mode != mode validation error.
            mode = self.initial_data["metrics"][i].get("mode")
            metric_key = metrickeys.get(name)
            item["metric_key"] = metric_key

            # Check that 'step' isn't set when sending Point mode metric values
            if step and ((metric_key and metric_key.mode != MetricMode.Series) or (mode and mode != "series")):
                errors[i]["step"].append("The step property cannot be set for Point type metrics.")
            # Check for incoherent mode and metric_key.mode
            if metric_key and mode and mode != str(metric_key.mode).lower():
                errors[i]["mode"].append(f"Metric Key {name} already exists and is of mode {metric_key.mode}.")
            # Only one metric value per point mode metric key
            if metric_key and metric_key.mode == MetricMode.Point and metric_key.metric_values_count > 0:
                errors[i]["__all__"].append(f"Metric Key {name} is of mode Point and already has a value.")
            # Only one metric value per step for series mode metric keys
            if step and metric_key and metric_key.mode == MetricMode.Series and metric_key.metric_values_count > 0:
                if metric_key.current_step_values > 0:
                    errors[i]["step"].append("This metric has already been set for this step.")
            # One metric key cannot have values both with and without step set
            if metric_key and metric_key.mode == MetricMode.Series and metric_key.metric_values_count > 0:
                if metric_key.values_with_steps > 0 and step is None:
                    errors[i]["step"].append(f"Cannot add a metric value without a step for metric key {metric_key.name} "
                                             "because it already has metric values with steps set.")
                elif metric_key.values_with_steps == 0 and step is not None:
                    errors[i]["step"].append(f"Cannot add a metric value with step set for metric key {metric_key.name} "
                                             "because it already has metric values with no step set.")

        if errors:
            raise ValidationError({"metrics": errors})

        return data

    def create(self, validated_data):
        # Create a list of metric keys to create
        metric_keys = []
        for item in validated_data["metrics"]:
            if not item.get("metric_key"):
                metric_key = MetricKey(
                    id=uuid.uuid4(),
                    name=item["name"],
                    model_version=validated_data["model_version"],
                    mode=item.get("mode")
                )
                metric_keys.append(metric_key)
                item["metric_key"] = metric_key

        # Create Metric Keys
        MetricKey.objects.bulk_create(metric_keys)

        # Create Metric Values
        MetricValue.objects.bulk_create(
            MetricValue(
                id=uuid.uuid4(),
                value=item["value"],
                metric=item["metric_key"],
                step=validated_data.get("step")
            )
            for item in validated_data["metrics"]
        )

        return validated_data


class DatasetSetSerializer(serializers.ModelSerializer):
    class Meta:
        model = DatasetSet
        fields = ("id", "name",)


class DatasetSerializer(serializers.ModelSerializer):
    state = EnumField(
        DatasetState,
        # Spectacular does not understand CreateOnlyDefault, so we document it in the help text too
        default=serializers.CreateOnlyDefault(DatasetState.Open),
        help_text=dedent("""
        Can only be updated through Ponos task authentication from a process containing this dataset
        and is limited to these transitions:

            Open → Building → Complete
                      ↕
                    Error

        When updated to `complete`, the dataset is automatically linked to the task used for authentication.

        When creating, this is always set to `open`.
        """),
    )

    creator = serializers.CharField(
        source="creator.display_name",
        read_only=True,
        default=serializers.CurrentUserDefault(),
        help_text="Display name of the user who created the dataset.",
    )

    set_names = serializers.ListField(child=serializers.CharField(max_length=50), write_only=True, default=["Training", "Validation", "Test"])
    sets = DatasetSetSerializer(many=True, read_only=True)

    # When creating the dataset, the dataset's corpus comes from the URL, so the APIView passes it through
    corpus = serializers.HiddenField(default=_corpus_from_context)

    # When creating the dataset, the dataset's creator is set to the current user automatically
    default_creator = serializers.HiddenField(
        source="creator",
        default=serializers.CreateOnlyDefault(serializers.CurrentUserDefault()),
    )
    set_elements = DatasetSetsCountField(
        help_text="Distribution of elements in sets. This value is set to null when listing multiple datasets.",
    )

    def validate_state(self, state):
        """
        Dataset's state update is limited to these transitions:
        Open → Building → Complete
                  ↕
                Error
        """
        # Ensure that only the open state can be set when creating
        if not isinstance(self.instance, Dataset):
            return DatasetState.Open

        if state == self.instance.state:
            return state

        transitions = {
            DatasetState.Open: (DatasetState.Building,),
            DatasetState.Building: (DatasetState.Complete, DatasetState.Error),
            DatasetState.Error: (DatasetState.Building,)
        }
        if state not in transitions.get(self.instance.state, ()):
            raise ValidationError(f"Transition from {self.instance.state} to {state} is not allowed.")
        return state

    def validate_set_names(self, set_names):
        if self.context["request"].method not in permissions.SAFE_METHODS and self.context["request"].method != "POST":
            raise ValidationError("This API endpoint does not allow updating a dataset's sets.")
        if set_names is not None and len(set(set_names)) != len(set_names):
            raise ValidationError("Set names must be unique.")
        return set_names

    def validate(self, data):
        data = super().validate(data)

        # No further validation needed when creating
        if not isinstance(self.instance, Dataset):
            return data

        state = data.get("state")
        if state is None or state == self.instance.state:
            return data

        # Dataset's state update requires a Ponos task authentication
        request = self.context.get("request")
        if request and not isinstance(request.auth, Task):
            raise ValidationError({
                "state": ["Ponos task authentication is required to update the state of a Dataset."]
            })
        # Dataset's state update is only allowed on tasks of Dataset processes, that have this dataset included
        if not request.auth.process.datasets.filter(id=self.instance.id).exists():
            raise ValidationError({"state": ["A task can only update the state of one of the datasets of its process."]})
        # Link a completed dataset to the current task which generated its artifacts
        if state == DatasetState.Complete:
            data["task_id"] = request.auth.id

        return data

    @transaction.atomic
    def create(self, validated_data):
        sets = validated_data.pop("set_names")
        dataset = Dataset.objects.create(**validated_data)
        DatasetSet.objects.bulk_create(
            DatasetSet(
                name=set_name,
                dataset_id=dataset.id
            ) for set_name in sets
        )
        return dataset

    class Meta:
        model = Dataset
        fields = (
            "id",
            "name",
            "description",
            "sets",
            "set_names",
            "set_elements",
            "state",
            # Only the corpus ID is actually serialized
            "corpus_id",
            # Hidden field retrieved from the API's context
            "corpus",
            "creator",
            # Hidden field to set the creator as the authenticated user
            "default_creator",
            "task_id",
            "created",
            "updated",
        )
        read_only_fields = (
            "id",
            "corpus_id",
            "corpus",
            "creator",
            "default_creator",
            "task_id",
            "created",
            "updated",
        )
        validators = [
            UniqueTogetherValidator(
                queryset=Dataset.objects.all(),
                fields=["corpus", "name"],
                message="A dataset with this name already exists in this corpus."
            )
        ]
        extra_kwargs = {
            "sets": {
                "error_messages": {
                    "empty": "Either do not specify set names to use the default values, or specify a non-empty list of names."
                }
            }
        }


class DatasetElementSerializer(serializers.ModelSerializer):
    element = ElementListSerializer(
        read_only=True,
        allow_null=False,
    )
    element_id = serializers.PrimaryKeyRelatedField(
        queryset=Element.objects.none(),
        style={"base_template": "input.html"},
        source="element",
        write_only=True,
    )
    dataset = serializers.HiddenField(
        default=_dataset_from_context,
        write_only=True,
    )

    class Meta:
        model = DatasetElement
        fields = ("set", "element", "element_id", "dataset")
        read_only_fields = ("element",)
        validators = [
            UniqueTogetherValidator(
                queryset=DatasetElement.objects.all(),
                fields=["dataset", "element_id", "set"],
                message="This element is already part of this set.",
            )
        ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if dataset := self.context.get("dataset"):
            self.fields["element_id"].queryset = Element.objects.filter(corpus=dataset.corpus)

    def validate_set(self, value):
        # The set must match the `sets` array defined at the dataset level
        dataset = self.context["dataset"]
        if dataset and value not in dataset.sets:
            raise ValidationError(f"This dataset has no set named {value}.")
        return value


class DatasetElementInfoSerializer(DatasetElementSerializer):
    dataset = serializers.PrimaryKeyRelatedField(
        queryset=Dataset.objects.none(),
        style={"base_template": "input.html"},
    )
    set = serializers.CharField(max_length=50)
    element = serializers.PrimaryKeyRelatedField(
        queryset=Element.objects.none(),
        style={"base_template": "input.html"},
    )

    class Meta(DatasetElementSerializer.Meta):
        fields = DatasetElementSerializer.Meta.fields + ("dataset",)


class ElementDatasetSerializer(serializers.ModelSerializer):
    dataset = DatasetSerializer()
    previous = serializers.UUIDField(allow_null=True, read_only=True)
    next = serializers.UUIDField(allow_null=True, read_only=True)

    class Meta:
        model = DatasetElement
        fields = ("dataset", "set", "previous", "next")
        read_only_fields = ("previous", "next")


class SelectionDatasetElementSerializer(serializers.Serializer):
    set_id = serializers.PrimaryKeyRelatedField(
        queryset=DatasetSet.objects.none(),
        source="set",
        write_only=True,
        help_text="UUID of a dataset set the elements will be added to.",
        style={"base_template": "input.html"},
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["set_id"].queryset = DatasetSet.objects.filter(
            dataset__corpus_id=self.context["corpus"].id
        ).select_related("dataset")

    def validate_set_id(self, set):
        if (
            (corpus := self.context.get("corpus"))
            and set.dataset.corpus_id != corpus.id
        ):
            raise ValidationError(f"Dataset {set.dataset.id} is not part of corpus {corpus.name}.")
        if set.dataset.state == DatasetState.Complete:
            raise ValidationError(f"Dataset {set.dataset.id} is marked as completed.")
        return set

    def create(self, validated_data):
        user = self.context["request"].user
        corpus = self.context["corpus"]

        DatasetElement.objects.bulk_create(
            (
                DatasetElement(element_id=elt_id, **validated_data)
                for elt_id in (
                    user.selected_elements
                    .filter(corpus=corpus)
                    .values_list("id", flat=True)
                )
            ),
            ignore_conflicts=True
        )
        return validated_data
